#! /usr/bin/env bash
##########################################################################
#                                                                        #
#  This file is part of Frama-C.                                         #
#                                                                        #
#  Copyright (C) 2007-2024                                               #
#    CEA (Commissariat à l'énergie atomique et aux énergies              #
#         alternatives)                                                  #
#                                                                        #
#  you can redistribute it and/or modify it under the terms of the GNU   #
#  Lesser General Public License as published by the Free Software       #
#  Foundation, version 2.1.                                              #
#                                                                        #
#  It is distributed in the hope that it will be useful,                 #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#  GNU Lesser General Public License for more details.                   #
#                                                                        #
#  See the GNU Lesser General Public License version 2.1                 #
#  for more details (enclosed in the file licenses/LGPLv2.1).            #
#                                                                        #
##########################################################################

# Note:
#
# While this script can be run locally, it is meant to run in the Frama-C CI.
# Thus, it expects to be run from the root of the Frama-C directory and that
# some CI artifacts are available. Namely:
#   - the 'website' as generated by the script Frama-C/dev/build-release.sh
# Availability of the files is NOT checked when the script starts.
# The following variables are used in the cloned website:
#   - GITLAB_USER_NAME (who triggered the job)
#   - GITLAB_USER_EMAIL
#   - GITLAB_USER_LOGIN
#   - GITLAB_USER_ID
# The script:
#   - clones the website as 'pub-website'
#   - checkout 'stable/release/<VERSION>-<LOWER CODENAME>'
#     if the branch does not exist in the repo, it is created
#   - copies the content of 'website' to 'pub-website'
#   - commits the changes
#   - pushes the branch
#     if the branch does not exist in the repo, a MR is created

##########################################################################

PUB_WEBSITE_GIT="git@git.frama-c.com:pub/pub.frama-c.com.git"
PUB_WEBSITE_DIR="pub-website"
SRC_WEBSITE_DIR="website"

VERSION="$(cat VERSION)"
VERSION_SAFE="$(cat VERSION | sed 's/~/-/')"
CODENAME="$(cat VERSION_CODENAME)"
LOWER_CODENAME="$(echo "$CODENAME" | tr '[:upper:]' '[:lower:]')"

BRANCH="release/$VERSION_SAFE-$LOWER_CODENAME"

echo "$FRAMA_CI_BOT_SSH_PRIVATE" | base64 -d > nix/frama-c-public/id_ed25519
chmod 400 nix/frama-c-public/id_ed25519

if ! git clone $PUB_WEBSITE_GIT $PUB_WEBSITE_DIR
then
  echo "Failed to clone website directory"
  exit 128
fi

cp -r $SRC_WEBSITE_DIR/* $PUB_WEBSITE_DIR
cd $PUB_WEBSITE_DIR

git lfs install
git config user.name  "Frama-CI Bot"
git config user.email "frama-ci-bot@frama-c.com"

function checkout {
  if git ls-remote --quiet --exit-code $PUB_WEBSITE_GIT $BRANCH
  then
    git checkout $BRANCH
  else
    git checkout -b $BRANCH
  fi
  if [ "$?" -ne "0" ]; then
    echo "Failed to checkout branch '$BRANCH', aborting"
    exit 128
  fi
}

function commit {
  git commit -F- <<EOF
[release] prepare $VERSION-$LOWER_CODENAME

On behalf of "$GITLAB_USER_NAME" <$GITLAB_USER_EMAIL> (@$GITLAB_USER_LOGIN)
EOF
  if [ "$?" -ne "0" ]; then
    echo "Failed to commit on branch"
    exit 128
  fi
}

function push {
  if git ls-remote --quiet --exit-code $PUB_WEBSITE_GIT $BRANCH
  then
    git push
  else
    git push \
      --set-upstream origin $BRANCH \
      -o merge_request.create \
      -o merge_request.title="Release $VERSION-$CODENAME" \
      -o merge_request.assign="$GITLAB_USER_ID"\
      -o merge_request.draft
  fi
  if [ "$?" -ne "0" ]; then
    echo "Failed to push branch"
    exit 128
  fi
}

checkout

git add -A

commit
push
